#!/usr/bin/env python
"""
Copyright 2015 ARC Centre of Excellence for Climate Systems Science

author: Scott Wales <scott.wales@unimelb.edu.au>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import print_function
from sqlalchemy import Column, Integer, String, Boolean, ForeignKey, Date, UniqueConstraint
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declarative_base
from ARCCSSive.data import *

import os
import glob

Base = declarative_base()

class Instance(Base):
    """
    A model variable from a specific run

    Search through these using :func:`ARCCSSive.CMIP5.Session.outputs()`

    .. attribute:: variable

        Variable name

    .. attribute:: experiment

        CMIP experiment

    .. attribute:: mip

        MIP table specifying output frequency and realm

    .. attribute:: model

        Model that generated the dataset

    .. attribute:: ensemble

        Ensemble member

    .. attribute:: realm

        Realm: ie atmos, ocean

    .. attribute:: versions

        List of :class:`Version` available for this output
    """
    __tablename__ = 'instances'
    id         = Column(Integer, name='instance_id', primary_key = True)

    variable   = Column(String, index=True)
    experiment = Column(String, index=True)
    mip        = Column(String, index=True)
    model      = Column(String, index=True)
    ensemble   = Column(String)
    realm      = Column(String)
    # Missing versions are labelled NA in database and v20110427 in drstree, this is CMOR documentation date
    # order doesn't work if version NA

    versions   = relationship('Version', order_by='Version.version', backref='variable')

    __table_args__ = (
            UniqueConstraint('variable','experiment','mip','model','ensemble'),
            )

    def latest(self):
        """
        Returns latest version/s available on raijin, first check in any version is_latest, then checks date stamp
        """
        if len(self.versions)==1: return self.versions 
        vlatest=[v for v in self.versions if v.is_latest]
        if vlatest==[]: 
            valid=[v for v in self.versions if v.version!="NA"]
            if valid==[]: return self.versions
            valid.sort(key=lambda x: x.version[-8:])
            vlatest.append(valid[-1])
            i=-2
            while i>=-len(valid) and valid[i].version[-8:]==vlatest[0].version[-8:]:
                vlatest.append(valid[i])
                i+=-1
        return vlatest
        
    def filenames(self):
        """
        Returns the file names from the latest version of this variable

        :returns: List of file names
        """
        return self.latest()[0].filenames()

    def drstree_path(self):
        """ 
        Returns the drstree path for this instance latest version 
        """
        drs_root="/g/data1/ua6/drstree/CMIP5/" # this should be passed as DRSTREE env var
        frequency=mip_dict[self.mip][0]
        return drs_root + "/".join([ self.model, 
                                     self.experiment,
                                     frequency,
                                     self.realm,
                                     self.variable,
                                     self.ensemble]) + "/latest" 

# Add alias to deprecated name
Variable = Instance

class Version(Base):
    """
    A version of a model run's variable

    .. attribute:: version

        Version identifier

    .. attribute:: path

        Path to the output directory

    .. attribute:: variable

        :class:`Variable` associated with this version

    .. attribute:: warnings

        List of :class:`VersionWarning` available for this output

    .. attribute:: files

        List of :class:`VersionFile` available for this output

    .. testsetup::

        >>> cmip5  = getfixture('session')

    >>> version = cmip5.query(Version).first()
    """
    __tablename__ = 'versions'
    id          = Column(Integer, name='version_id', primary_key = True)
    instance_id = Column(Integer, ForeignKey('instances.instance_id'), index=True)

    version     = Column(String)
    path        = Column(String)
    dataset_id  = Column(String)
    is_latest   = Column(Boolean)
    checked_on  = Column(String)
    to_update   = Column(Boolean)

    warnings   = relationship('VersionWarning', order_by='VersionWarning.id', 
                              backref='version', cascade="all, delete-orphan", passive_deletes=True)
    files   = relationship('VersionFile', order_by='VersionFile.id', 
                            backref='version', cascade="all, delete-orphan", passive_deletes=True)


    def glob(self):
        """
        Get the glob string matching the CMIP5 filename

        .. testsetup::

            >>> import six
            >>> cmip5  = getfixture('session')
            >>> version = cmip5.query(Version).first()

        >>> six.print_(version.glob())
        a_6hrLev_c_d_e*.nc
        """
        return '%s_%s_%s_%s_%s*.nc'%(
            self.variable.variable,
            self.variable.mip,
            self.variable.model,
            self.variable.experiment,
            self.variable.ensemble)

    def build_filepaths(self):
        """
        Returns the list of files matching this version

        :returns: List of file names

        .. testsetup::

            >>> cmip5  = getfixture('session')
            >>> version = cmip5.query(Version).first()

        >>> version.build_filepaths()
        []
        """
        g = os.path.join(self.path, self.glob())
        return glob.glob(g)
         
    def filenames(self):
        """
        Returns the list of filenames for this version

        :returns: List of file names

        .. testsetup::

            >>> cmip5  = getfixture('session')
            >>> version = cmip5.query(Version).first()

        >>> version.filenames()
        []
        """
        return [x.filename for x in self.files] 
         
    def tracking_ids(self):
        """
        Returns the list of tracking_ids for files in this version

        :returns: List of tracking_ids

        .. testsetup::

            >>> cmip5  = getfixture('session')
            >>> version = cmip5.query(Version).first()

        >>> version.tracking_ids()
        []
        """
        return [x.tracking_id for x in self.files] 

    def drstree_path(self):
        """ 
        Returns the drstree path for this particular version 
        """
        if self.version!='NA':
            version=self.version
        else:
            version='v20110427'
        return self.variable.drstree_path().replace('latest',version)
        
class VersionWarning(Base):
    """
    Warnings associated with a output version
    """
    __tablename__ = 'warnings'

    id         = Column(Integer, name='warning_id', primary_key = True)
    warning    = Column(String)
    added_by   = Column(String)
    added_on   = Column(Date)
    version_id = Column(Integer, ForeignKey('versions.version_id'), index=True)

    def __str__(self):
        return u'%s (%s): %s'%(self.added_on, self.added_by, self.warning) 


class VersionFile(Base):
    """
    Files associated with a output version
    """
    __tablename__ = 'files'

    id           = Column(Integer, name='file_id', primary_key = True)
    filename     = Column(String)
    tracking_id  = Column(String)
    md5          = Column(String)
    sha256       = Column(String)
    version_id   = Column(Integer, ForeignKey('versions.version_id'), index = True)

    def __str__(self):
        return '%s'%(self.filename)



# wrapper functions for py-esgf-search
#!/usr/bin/env python
"""
Copyright 2016 ARC Centre of Excellence for Climate Systems Science

author: Paola Petrelli <paola.petrelli@utas.edu.au>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import print_function

from pyesgf.logon import LogonManager
from pyesgf.search import SearchConnection
from pyesgf.search.results import DatasetResult as DatasetResult
from pyesgf.search.results import FileResult as FileResult
from ARCCSSive.data import model_names_dict

import sys

def logon(openid, password):
    ''' Login on ESGF with credentials, returns connection object '''
    lm=LogonManager()
    try:
        lm.logon_with_openid(openid, password, bootstrap=True, update_trustroots=True)
    except:
        e = sys.exc_info()[1]
        print("Logon Error: ",  e)
    return lm 

def logoff(lm):
    ''' Logoff ESGF, returns True if successful '''
    lm.logoff()
    return not lm.is_logged_on()

class ESGFSearch(object):
    ''' defines a ESGF search object 
        :param connection: The SearchConnection
        :param constraints: A dictionary of initial constraints
        :param search_type: One of TYPE_* constants defining the document
            type to search for.  Overrides SearchContext.DEFAULT_SEARCH_TYPE
        :param facets: The list of facets for which counts will be retrieved
            and constraints be validated against.  Or None to represent all facets.
        :param fields: A list of field names to return in search responses
        :param replica: A boolean defining whether to return master records
            or replicas, or None to return both.
        :param latest: A boolean defining whether to return only latest verisons
            or only non-latest versions, or None to return both.
        :param shards: list of shards to restrict searches to.  Should be from the list
            self.connection.get_shard_list()
        :param from_timestamp: Date-time string to specify start of search range 
            (e.g. "2000-01-01T00:00:00Z"). 
        :param to_timestamp: Date-time string to specify end of search range
            (e.g. "2100-12-31T23:59:59Z")
        **kwargs for all the functions are optional constraints as defined in the ESGF RESTful Search API 
        http://
    '''

    def search_node(self, **kwargs):
        ''' Opens search connection and creates search context object self.ctx 
            by default searches CMIP5 latest version, not replica
        :argument node: primary node to search 
                        default "http://pcmdi.llnl.gov/esg-search" 
        :argument distrib: default True search across all nodes 
                           both these arguments can be passed as part of **kwargs
        :argument **kwargs: optional constraints to apply, listed in class comment
        :return: 
        ''' 
        node, distrib, replica = ["http://pcmdi.llnl.gov/esg-search",True, False]
        if "node" in kwargs.keys(): node = kwargs.pop('node')
        if "distrib" in kwargs.keys(): distrib = kwargs.pop('distrib')
        if "replica" in kwargs.keys(): replica = kwargs.pop('replica')
        if 'model' in kwargs.keys():
            kwargs['model']=self.model_names(kwargs['model']) 
        self.conn = SearchConnection(node, distrib=distrib)
        self.ctx = self.conn.new_context(project='CMIP5', latest=True, 
                                           replica=replica, **kwargs)
        return 

    def model_names(self, model):
        ''' Returns model with valid name for ESGF search api
        :argument model: string model name passed as constraint
        :return: a string of correct model name where two names exist for same model
        ''' 
        if model in model_names_dict.keys():
            return  model_names_dict[model]
        else:
            return model

    def get_ds(self, **kwargs):
        ''' Returns list of dataset objects selected by search_node, further constraints can be applied  
        :argument **kwargs: optional constraints to apply, listed in class comment
        :return: A list of pyESGF DatasetResult objects
        '''
        return self.ctx.search(**kwargs)

    def ds_filter(self, **kwargs):
        '''Narrows down search results 
        :argument **kwargs: optional constraints to apply, listed in class comment
        :return: A list of pyESGF DatasetResult objects
        '''
        return self.ctx.constrain(**kwargs)

    def which_facets(self, *args):
        ''' Lists available facets, applies first additional constraints if any
        :argument **kwargs: optional constraints to apply, listed in class comment
        :return: A list of available facets ie any parameter key,value pair returned by search
        ''' 
        # this needs fixing leave it aside for moment
        print(dir(self.ctx))
        return self.ctx.facets(*args)

    def facet_values(self, *args): 
        ''' return available values for a particular facet as currently constrained 
        :argument *args: optional facet (constraints key), listed in class comment
        :return: A list of available values for the input facet/s in current search result
        ''' 
        return self.ctx.facets_count(*args)

    def facet_options(self): 
        ''' return facets available to narrow down further search results
        ''' 
        return self.ctx.get_facet_options()

    #def which_fields(self, **kwargs):
    #    ''' Lists available facets, applies first additional constraints if any
    #    :argument **kwargs: optional constraints to apply, listed in class comment
    #    :return: A list of available facets ie any parameter key,value pair returned by search
    #    ''' 
    #    return self.ctx.fields()

    def ds_count(self): 
        ''' return number of datasets in search result '''
        return self.ctx.hit_count

    def ds_ids(self): 
        ''' return list of dataset_id for datasets in search result '''
        return [ x.dataset_id for x in self.ctx.search()]
      
    def ds_versions(self):
        ''' return list of versions for datasets in search result '''
        return [ 'v'+x.json['version'] for x in self.ctx.search()]

    def ds_variables(self):
        ''' return set of variables for datasets in search result '''
        var_list=[ x.json['cf_standard_name'] for x in self.ctx.search()]
        return set( x for i in range(len(var_list)) for x in var_list[i])


# methods to extend classes DatasetResult and FileResult
def variables(self):
    ''' return list of variables for dataset  '''
    self.variables=self.json['cf_standard_name']
    return

def files(self):
    ''' return list of FileResult for one  dataset object '''
    return self.file_context().search()

def filenames(self):
    ''' return list of filenames for files in the dataset object '''
    return [x.filename for x in self.files()]

def tracking_ids(self):
    ''' return list of tracking_ids for files in the dataset object '''
    return [x.tracking_id for x in self.files()]

def checksums(self):
    ''' return list of checksums for files in the dataset object '''
    return [x.checksum for x in self.files()]

def chksum_type(self):
    ''' return checksum_type for files in the dataset object '''
    return self.files()[0].checksum_type

# all FileResult properties
# 'checksum', 'checksum_type', 'context', 'download_url', 'file_id', 'filename', 'index_node', 'json', 'las_url', 'opendap_url', 'size', 'tracking_id', 'urls']

def list_attributes(self):
    ''' return a listed of available attributes of Dataset/FileResult '''
    return self.json.keys()

def get_attribute(self, attr):
    ''' return a list for the attribute of Dataset/FileResult specified as input '''
    return self.json[attr]

# Adding methods to DatasetResult class
DatasetResult.variables = variables
DatasetResult.files = files
DatasetResult.filenames = filenames
DatasetResult.tracking_ids = tracking_ids
DatasetResult.checksums = checksums
DatasetResult.chksum_type = chksum_type
DatasetResult.list_attributes = list_attributes
DatasetResult.get_attribute = get_attribute
# Adding methods to FileResult class
FileResult.list_attributes = list_attributes
FileResult.get_attribute = get_attribute



#!/usr/bin/env python
# Copyright 2016 ARC Centre of Excellence for Climate Systems Science
# author: Scott Wales <scott.wales@unimelb.edu.au>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import print_function
import socket
import sys
import os
from pprint import pformat
from .pkg_info import __version__

class Info:
    def __init__(self):
        self.host= socket.getfqdn(),
        self.version= __version__,
        self.py_version= sys.version,
        self.databases= {
            'CMIP5': os.environ.get('CMIP5_DB',None),
            }

    def __str__(self):
        return pformat(vars(self))

info = Info()



# This collects all functions to update database using SQLalchemy
#!/usr/bin/env python
"""
Copyright 2016 ARC Centre of Excellence for Climate Systems Science

author: Paola Petrelli <paola.petrelli@utas.edu.au>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import print_function

from sqlalchemy.orm.exc import NoResultFound
#from ARCCSSive.CMIP5.Model import Instance, Version, VersionFile, VersionWarning


def search_item(db, klass, **kwargs):
    """
    Search for item in DB, if it can't be found return False
    Otherwise returns only first item
    """
    try:
        item = db.query(klass).filter_by(**kwargs).one()
    except NoResultFound:
        return False
    return item

def insert_unique(db, klass, **kwargs):
    """
    Insert an item into the DB if it can't be found
    NB need to commit before terminating session
    Returns id, True if new item, id False if existing
    """
    item = search_item(db, klass, **kwargs)
    new=False
    if not item:
        item = klass(**kwargs)
        db.add(item)
        new=True
        db.commit() 
    return item, new

def add_bulk_items(db, klass, rows):
    """Batched INSERT statements via the ORM "bulk", using dictionaries.
       input: rows is a list of dictionaries """
    db.bulk_insert_mappings(klass,rows) 
    db.commit()
    return

def update_item(db, klass, item_id, newvalues):
    '''Update database item 
       :argument: db database SQLalchemy connection session 
       :argument: klass class representing table objects
       :argument: item_id the id for row to update
       :argument: newvalues a dictionary with the column keys, values to be updated
       :return nothing
    '''
    db.query(klass).filter_by(id=item_id).update(newvalues)
    db.commit()
    return

def commit_changes(db):
    ''' Commit changes to database '''
    return db.commit()




#!/usr/bin/env python
# Copyright 2016 ARC Centre of Excellence for Climate Systems Science
# author: Scott Wales <scott.wales@unimelb.edu.au>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import print_function

import pkg_resources
__name__ = 'ARCCSSive'
__version__ = pkg_resources.get_distribution(__name__).version



#!/usr/bin/env python
"""
Copyright 2016 ARC Centre of Excellence for Climate Systems Science

author: Scott Wales <scott.wales@unimelb.edu.au>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import pickle

# load mip and frequency dictionaries
def _read_pickle():
    # Read the packaged data
    from pkg_resources import resource_stream
    pickle_stream = resource_stream(__name__, 'data/cmip_dict_pickle')

    mip_dict = pickle.load(pickle_stream)
    frq_dict = pickle.load(pickle_stream)
    model_names_dict = pickle.load(pickle_stream)
    return mip_dict, frq_dict, model_names_dict

mip_dict, frq_dict, model_names_dict = _read_pickle()



# This collects all other helper functions 
#!/usr/bin/env python
"""
Copyright 2016 ARC Centre of Excellence for Climate Systems Science

author: Paola Petrelli <paola.petrelli@utas.edu.au>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import print_function

import os
import itertools
from datetime import date
import glob
import subprocess
import re
#import cdms2 to open netcdf files
from collections import defaultdict
from ARCCSSive.data import mip_dict, frq_dict 
from ARCCSSive.CMIP5.update_db_functions import add_bulk_items, update_item 
from ARCCSSive.CMIP5.Model import Instance, VersionFile


def combine_constraints(**kwargs):
    ''' Works out any possible combination given lists of constraints
        :argument dictionary, keys are fields and values are lists of values for each field
        :return: a set of dictionaries, one for each constraints combination i
    '''
    try:
        return [dict(itertools.izip(kwargs, x)) for x in itertools.product(*kwargs.itervalues())]
    except:
        return [dict(zip(kwargs, x)) for x in itertools.product(*kwargs.values())]

def join_varmip(var0,mip0):
    ''' List all combinations of selected variables and mips '''
    comb = ["_".join(x) for x in itertools.product(*[var0,mip0])]
    print(comb)
    return comb

def get_instance(dataset_id):
    ''' Break dataset_id from ESGF search in dictionary of instance attributes '''
    bits=dataset_id.split(".")
    return {'model': bits[3],'experiment': bits[4],'realm':bits[6],'mip':bits[7],'ensemble':bits[8],'version':bits[9].split("|")[0]}

def compare_instances(db,remote,local,const_keys,admin):
    ''' Compare remote and local search results they're both a list of dictionaries
        :argument db: sqlalchemy local db session 
        :argument remote: each dict has keys version, files (objs), filenames, tracking_ids, dataset_id 
        :argument local:  list of version objects
        :argument const_keys:  list of instance attributes defined by user constraints
        :argument admin:  boolean if True user is admin and can update db directly, optherwise new info saved in log file
        :return: remote, local with updated attributes 
    '''
    global logfile
    logdir="/g/data1/ua6/unofficial-ESG-replica/tmp/pxp581/requests/"
    if not admin:
        logfile=logdir+"log_" + os.environ['USER'] + "_" + today.replace("-","") + ".txt"
        print(logfile)
    # a list of the unique constraints defining one instance in the database which are not in user constraints
    undefined=[x for x in Instance.__table_args__[0].columns.keys() if x not in const_keys]
    # loop through all returned remote datasets
    for ind,ds in enumerate(remote):
        # loop through all local versions
        ds_instance=get_instance(ds['dataset_id'])
        for v in local:
            dummy=[False for key in undefined if  ds_instance[key] != v.variable.__dict__[key]]
            if False in dummy:
                continue
            v.checked_on = today
            # compare files for all cases except if version regular but different from remote 
            if v.version in [ds['version'],'NA',r've\d*']:
                extra = compare_files(db,ds,v,admin)
                # if tracking_ids or checksums are same
                if extra==set([]):
                    v.to_update = False
                else:
                    v.to_update = True
                    if not admin: 
                        ds_info=[str(x) for x in ds_instance.items()]
                        write_log(" ".join(["update"]+ds_info+[v.version,v.path,"\n"]))
            # if local dataset_id is the same as remote skip all other checks
            if v.dataset_id==ds['dataset_id']:
                v.is_latest = True
            # if version same as latest on esgf 
            elif v.version == ds['version']:
                v.dataset_id = ds['dataset_id']
                v.is_latest = True
            # if version undefined 
            elif v.version in ['NA',r've\d*']:
                if extra==set([]):
                    v.version = ds['version']
                    v.dataset_id = ds['dataset_id']
                    v.is_latest = True
            # if version different or undefined but one or more tracking_ids are different
            # assume different version from latest
            # NB what would happen if we fix faulty files? tracking_ids will be same but md5 different, 
            # need to set a standard warning for them
            else:
                v.is_latest = False
                if v.version > ds['version']: 
                    print("Warning!!!")
                    print(" ".join(["Local version",v.version,"is more recent than the latest version",ds['version'], "found on ESGF"]))
                if v.dataset_id is None: v.dataset_id = "NA"
    # update local version on database
            if admin: 
                db.commit()
            else:
                if db.dirty:
                    line=["version"]+ds_instance.values()[:-1]+[v.version,str(v.id),str(v.dataset_id),
                          str(v.is_latest),str(v.checked_on),"\n"]
                    write_log(" ".join(line))
                
    # add to remote dictionary list of local identical versions
        remote[ind]['same_as']=[v.id for v in local if v.dataset_id == ds['dataset_id']] 
    try:
        flog.close()
    except:
        pass
    return remote, local

def compare_files(db,rds,v,admin):
    ''' Compare files of remote and local version of a dataset
        :argument rds: dictionary of remote dataset object selected attributes  
        :argument v:  local version object   
        :return: result set, NB updating VerisonFiles object in databse if calculating checksums 
    '''
    extra=None
    local_files_num=len(v.files)
    # if there are no files on db for local version add them
    if v.files==[]:
        rows=[]
        for f in v.build_filepaths():
            rows.append(dict(filename=f.split("/")[-1], version_id=v.id))
        if admin:   
            add_bulk_items(db, VersionFile, rows)
        else:
            for r in rows:
                write_log("new file "+ str(r) + "\n")
        local_files_num=len(rows)
    # first compare tracking_ids if all are present in local version
    # if a file is INVALID or missing skip both tracking-id and checksums comparison
    local_ids=[x for x in v.tracking_ids() if x not in [None,""]]
    if "INVALID" not in local_ids or local_files_num != len(rds['files']):
        return extra
    if len(local_ids)>0:
        extra = compare_tracking_ids(rds['tracking_ids'],local_ids)
    # calculate checksums and update local db if necessary  
    # uncomment this to check also if tracking_ids are the same
    #if extra is None or  extra==set([]):
    # if tracking_ids are not present compare checksums
    if extra is None:
        local_sums=[]
        cktype=str(rds['checksum_type']).lower()
        for f in v.files:
            try:
                cksum=f.__dict__[cktype] 
            except (TypeError, KeyError):
                #print("type or key error ",cktype)
                cksum=None
            if cksum in ["",None]:
                cksum=check_hash(v.path+"/"+f.filename,cktype)
                if admin: 
                    update_item(db,VersionFile,f.id,{cktype:cksum})
                else:
                    write_log(" ".join([cktype,str(f.id),cksum,"\n"]))
            local_sums.append(cksum) 
        extra = compare_checksums(rds['checksums'],local_sums)
    return extra 
            
def compare_tracking_ids(remote_ids,local_ids):
    ''' Compare the lists of the tracking_ids from a remote and a local version of a dataset
        :argument remote_ids: list of remote tracking_ids  
        :argument local_ids: list of local tracking_ids  
        :return: result set 
    '''
    return set(remote_ids).symmetric_difference(local_ids)
    
def compare_checksums(remote_sums,local_sums):
    ''' Compare the lists of the checksums from a remote and a local version of a dataset
        :argument remote_sums: list of remote checksums  
        :argument local_sums: list of local checksums  
        :return: result set 
    '''
    return set(remote_sums).symmetric_difference(local_sums)
    
# these functions are to manage drstree and tmp/tree directories

def list_drstree(**kwargs):
    ''' find directories matching kwargs constraints in drstree
        check if they are in database,
        if not add them to db
        return: list of matches '''
    if 'mip' in kwargs.keys(): kwargs['frequency']=frequency(kwargs['mip'])
    indir=drstree + drs_glob(**kwargs)
    return glob.glob(indir)

def list_tmpdir(flist):
    ''' this read from file list of instances on tmp/tree and return the ones matching constraints '''
# skip first line
# variable,mip_table,model,experiment,ensemble,realm,version,path
    keys=['variable','mip','model','experiment','ensemble','realm','version','path']
    f=open(flist,'r')
    inst_list=[]
    lines=f.readlines()
    for line in lines[1:]:
        values=line[:-1].split(',')
        inst_list.append(dict(zip(keys, values)))
    f.close()
    return inst_list

def list_logfile(flist):
    ''' this read from file list of instances from download log file and return the ones matching constraints '''
    keys=['variable','mip','model','experiment','ensemble','realm','version','path','cks_type','files']
    file_keys=['filename','tracking_id','checksum']
    f=open(flist,'r')
    inst_list=[]
    file_list=[]
    lines=f.readlines()
    values=lines[0].replace("\n","").split(',')
    var=values[0]
    ds_dict=dict(zip(keys[:-1], values))
    for line in lines[1:]:
        values=line.replace("\n","")[:-1].split(',')
        if values[0]!=var:
            file_list.append(dict(zip(file_keys, values)))
        else:
            ds_dict['files']=file_list
            inst_list.append(ds_dict)
            file_list=[]
            ds_dict=dict(zip(keys[:-1], values))
    f.close()
    return inst_list

def file_glob(**kwargs):
    """ Get the glob string matching the CMIP5 filename
    """
    value=defaultdict(lambda: "*")
    value.update(kwargs)
    return '%s_%s_%s_%s_%s*.nc'%(
            value['variable'],
            value[ 'mip'],
            value['model'],
            value['experiment'],
            value['ensemble'])

def drs_glob(**kwargs):
    """ Get the glob string matching the directory structure under drstree
    """
    value=defaultdict(lambda: "*")
    value.update(kwargs)
    return '%s/%s/%s/%s/%s/%s/%s'%(
        value['model'],
        value['experiment'],
        value['frequency'],
        value['realm'],
        value['variable'],
        value['ensemble'],
        value['version'])

def tree_glob(**kwargs):
    """ Get the glob string matching the directory structure under tmp/tree
    """
    value=defaultdict(lambda: "*")
    value.update(kwargs)
    return '%s/%s/%s/%s/%s/%s'%(
        value['model'],
        value['experiment'],
        value['frequency'],
        value['realm'],
        value['variable'],
        value['ensemble'])

def drs_details(path):
    ''' Split the drstree path in model, experiment, frequency, realm, variable, ensemble '''
    keys=['model','experiment', 'frequency', 'realm', 'variable','ensemble']
    values=path.replace(drstree,"").split('/')
    dummy=dict((keys[i],values[i]) for i in range(len(values)))
    return dummy.pop('frequency'), dummy

def file_details(fname):
    ''' Split the filename in variable, MIP code, model, experiment, ensemble (period is excluded) '''
    keys=['variable','mip','model','experiment','ensemble']
    values = fname.split('_')
    if len(values) >= 5:
        return dict((keys[i],values[i]) for i  in range(len(values[:-1])))
    else:
        return 

def find_version(bits,string):
    ''' Returns matching string if found in directory structure '''
    dummy = filter(lambda el: re.findall( string, el), bits)
    if len(dummy) == 0:
        return 'not_specified'
    else:
        return dummy[0]

#def get_trackid(fpath):
#    ''' Return trackid from netcdf file '''
#    # open netcdf file
#    try:
#        f = cdms2.open(fpath,'r')
#    except:
#        print("INVALID NETCDF,%s" % fpath)
#        return "INVALID" 
#    # read tracking id attribute
#    try:
#        trackid=f.tracking_id
#    except:
#        trackid=None
#    # close file
#    f.close()
#    return trackid

def list_drs_versions(path):
    ''' Returns matching string if found in directory structure '''
    return [x.split("/")[-1] for x in glob.glob(path+"/v*")]

def list_drs_files(path):
    ''' Returns matching string if found in directory structure '''
    return [x.split("/")[-1] for x in glob.glob(path+"/*.nc")]
 
def get_mip(path):
    ''' Returns mip for instance 
        input: instance path
    '''
    onefile = os.path.basename(glob.glob(path + "/latest/*.nc")[0]) 
    dummy = file_details(onefile)
    return dummy['mip']

def tree_path(drspath):
    ''' Returns the tmp/tree path for a particular instance &  version from 
        input: drstree path for one the version files
    '''
    path=os.path.realpath(drspath)
    return "/".join(path.split("/")[:-1])

def check_hash(path,hash_type):
    ''' Execute md5sum/sha256sum on file on tree and return checksum value '''
    hash_cmd="md5sum"
    if hash_type in ["SHA256","sha256"]: hash_cmd="sha256sum"
    try:
        return subprocess.check_output([hash_cmd, path]).split()[0]
    except:
        print("Warning cannot calculate ",hash_type," for file ",path)
        return ""

# functions to manage dictionaries
def assign_mips(**kwargs):
    ''' Append the cmip5 mip tables corresponding to the input frequency 
        return updates list of mips '''
    if kwargs['mip'] is None: kwargs['mip']=[]
    if kwargs['frq']:
        kwargs['mip'].extend([y for x in kwargs['frq'] for y in frq_dict[x]]) 
    return list(set([x for x in kwargs['mip']]))

def frequency(mip):
    ''' returns frequency for input mip '''
    return  mip_dict[mip]

# functions to write logs
def write_log(line):
    ''' add str to log file, open new file if does not exist yet '''
    global flog
    try:
        flog.write(line)
    except:
        flog=open(logfile,"a")
        flog.write(line)
    return
      
      
     
# this should be taken by setting environment variable DRSTREE
# define root cirectory for drstree and /tmp/tree
try:
    drstree = os.environ['DRSTREE']
except KeyError:
    drstree = "/g/data1/ua6/drstree/CMIP5/GCM/"
drstree="/g/data1/ua8/cmip-download/drstree/CMIP5/GCM/"
tmptree="/g/data1/ua6/unofficial-ESG-replica/tmp/tree/"

# define date string for current date
today = date.today().strftime('%Y-%m-%d')




#!/usr/bin/env python
"""
Copyright 2015 ARC Centre of Excellence for Climate Systems Science

author: Scott Wales <scott.wales@unimelb.edu.au>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import print_function

import os

from sqlalchemy import create_engine, func, select, and_
from sqlalchemy.orm import sessionmaker

from ARCCSSive.CMIP5.Model import Base, Instance

SQASession = sessionmaker()

class Session(object):
    """Holds a connection to the catalog

    Create using :func:`ARCCSSive.CMIP5.connect()`
    """

    def query(self, *args, **kwargs):
        """Query the CMIP5 catalog

        Allows you to filter the full list of CMIP5 outputs using `SQLAlchemy commands <http://docs.sqlalchemy.org/en/rel_1_0/orm/tutorial.html#querying>`_

        :return: A SQLalchemy query object
        """
        return self.session.query(*args, **kwargs)

    def files(self, **kwargs):
        """ Query the list of files

        Returns a list of files that match the arguments

        :argument **kwargs: Match any attribute in :class:`Model.Instance`, e.g. `model = 'ACCESS1-3'`

        :return: An iterable returning :py:class:`Model.File`
            matching the search query
        """
        raise NotImplementedError

    def models(self):
        """ Get the list of all models in the dataset

        :return: A list of strings
        """
        return [x[0] for x in self.query(Instance.model).distinct().all()]

    def experiments(self):
        """ Get the list of all experiments in the dataset

        :return: A list of strings
        """
        return [x[0] for x in self.query(Instance.experiment).distinct().all()]

    def variables(self):
        """ Get the list of all variables in the dataset

        :return: A list of strings
        """
        return [x[0] for x in self.query(Instance.variable).distinct().all()]

    def mips(self):
        """ Get the list of all MIP tables in the dataset

        :return: A list of strings
        """
        return [x[0] for x in self.query(Instance.mip).distinct().all()]

    def outputs(self, **kwargs):
        """ Get the most recent instances matching a query

        Arguments are optional, using them will select only matching outputs

        :argument variable: CMIP variable name
        :argument experiment: CMIP experiment
        :argument mip: MIP table
        :argument model: Model used to generate the dataset
        :argument ensemble: Ensemble member

        :return: An iterable sequence of :class:`ARCCSSive.CMIP5.Model.Instance`
        """
        return self.query(Instance).filter_by(**kwargs)


def connect(path = None):
    """Connect to the CMIP5 catalog

    :return: A new :py:class:`Session`

    Example::

    >>> from ARCCSSive import CMIP5 
    >>> cmip5   = CMIP5.DB.connect() # doctest: +SKIP
    >>> outputs = cmip5.query() # doctest: +SKIP
    """

    if path is None:
        try:
            # Get the path from the environment
            path = os.environ['CMIP5_DB']
        except KeyError:
            raise Exception('Environment variable $CMIP5_DB should point to the database')

    engine = create_engine(path)
    Base.metadata.create_all(engine)
    SQASession.configure(bind=engine)

    connection = Session()
    connection.session = SQASession()
    return connection

